from symMaps.base import *
_adjF = adj.rc_pd


class SimpleSys:
	def __init__(self, db = None, v = None, maps = None, iterAux = True):
		self.db = noneInit(db, SimpleDB()) # database
		self.v = noneInit(v, {}) # variables
		self.maps = noneInit(maps, {}) # mapping from pandas index to global linear variable index.
		self.auxMaps = {}
		self.auxMapsIdx = {}
		self.iterAux = iterAux

	def x0(self, attr = 'v', fill_value = 0):
		return np.concatenate([self.xFromDb(name, attr = attr, fill_value = fill_value) for name in self.v])

	def xFromDb(self, name, attr = 'v', fill_value = 0):
		return self.db[name].array(attr = attr) if name in self.db.symbols else np.full(len(self.v[name]), fill_value)

	def compile(self):
		keys, vals = list(self.v.keys()), list(self.v.values())
		steps = np.array([0]+[1 if x is None else len(x) for x in vals]).cumsum()
		self.len = steps[-1]
		self.maps = {keys[i]: pd.Series(range(steps[i], steps[i+1]), index = getIndex(vals[i])) for i in range(len(keys))}

	def __getitem__(self, item):
		return (self.auxMaps | self.maps)[item]

	def __iter__(self):
		return iter(self.maps | self.auxMaps) if self.iterAux else iter(self.maps)

	def __len__(self):
		return len(self.maps | self.auxMaps) if self.iterAux else len(self.maps)

	def __call__(self, x, name, **kwargs):
		""" Subset x with index of variable 'name' using linear indexing from self.maps."""
		return x[self[name]]

	def get(self, x, name, **kwargs):
		""" __call__ method, but returned as pandas object"""
		return x[self[name][0]] if self[name].index is None else pd.Series(x[self[name]], index = self[name].index, name = name)
		
	def getr(self, x, name, **kwargs):
		""" Like the get method, but more robust (adds more potential adjustments to the symbol) """
		if self.symbols[name] is None: # check if the symbol is a scalar
			return x[self[name][0]] # return without pandas object
		else:
			k = _adjF(self[name], **kwargs)
			return pd.Series(x[k], index = k.index, name = name)

	def unloadSol(self, x):
		return {k: self.get(x, k) for k in self}

	# def applyMapGlobalIdx(self, symbol, m):
	# 	return pd.Series(symbol[m.values].values, m.index)

	# def addSymFromMap(self, name, symbol, m):
	# 	self.auxMapsIdx[name] = m
	# 	self.auxMaps[name] = self.applyMapGlobalIdx(self[symbol], m)

	# def _dropna(self, glbSymbol, m):
	# 	""" Only keep elements in mapping (pd.Series) that are in the "global" version in self"""
	# 	return self._reverseMap(adj.rc_pd(self._reverseMap(m), self[glbSymbol]))

	# @staticmethod
	# def _reverseMap(m):
	# 	return pd.Series(m.index.values, index = pd.MultiIndex.from_tuples(m.values, names = m.index.names)
	# 														 if isinstance(m.index, pd.MultiIndex) else pd.Index(m.values, name = m.index.name))

	# def addLaggedSym(self, name, symbol, lags, dropna = False, **kwargs):
	# 	m = self.lagMaps(adj.rc_pd(self[symbol], **kwargs), lags)
	# 	return self.addSymFromMap(name, symbol, self._dropna(symbol, m) if dropna else m)

	# def getLagFromSol(self, x, symbol, lags, dropna=False, **kwargs):
	# 	""" Return a lagged symbol like self.addLaggedSym, but without adding it to the compilation stage."""
	# 	glbIdx = self.applyMapGlobalIdx(self[symbol], self.lagMaps(adj.rc_pd(self[symbol], **kwargs), lags))
	# 	return pd.Series(x[glbIdx], index = glbIdx.index)

	# def getLag(self, x, lags, dropna = False, **kwargs):
	# 	""" Return a lagged symbol like getLagFromSol, but where x is the "non-shifted" symbol (pd.Series) instead of the global vector"""
	# 	m = self.lagMaps(adj.rc_pd(x, **kwargs), lags)
	# 	return pd.Series(x[m.values].values, index = x.index)

	# def addRolledSym(self, name, symbol, rolls, **kwargs):
	# 	self.addSymFromMap(name, symbol, self.rollMaps(adj.rc_pd(self[symbol], **kwargs), rolls))

	# def getRollFromSol(self, x, symbol, rolls, dropna=False, **kwargs):
	# 	""" Return a rolled symbol like self.addRolledSym, but without adding it to the compilation stage."""
	# 	glbIdx = self.applyMapGlobalIdx(self[symbol], self.rollMaps(adj.rc_pd(self[symbol], **kwargs), rolls))
	# 	return pd.Series(x[glbIdx], index = glbIdx.index)

	# def getRoll(self, x, rolls, dropna = False, **kwargs):
	# 	""" Return a shifted symbol like getShiftFromSol, but where x is the "non-shifted" symbol (pd.Series) instead of the global vector"""
	# 	m = self.rollMaps(adj.rc_pd(x, **kwargs), rolls)
	# 	return pd.Series(x[m.values].values, index = x.index)

	# def addShiftedSym(self, name, symbol, shifts, dropna = False, opt = None, **kwargs):
	# 	m = self.shiftMaps(adj.rc_pd(self[symbol], **kwargs), shifts, **noneInit(opt, {}))
	# 	return self.addSymFromMap(name, symbol, self._dropna(symbol, m) if dropna else m)

	# def getShiftFromSol(self, x, symbol, shifts, dropna=False, opt= None, **kwargs):
	# 	""" Return a shifted symbol like self.addShiftedSym, but without adding it to the compilation stage."""
	# 	glbIdx = self.applyMapGlobalIdx(self[symbol], self.shiftMaps(adj.rc_pd(self[symbol], **kwargs), shifts, **noneInit(opt, {})))
	# 	return pd.Series(x[glbIdx], index = glbIdx.index)

	# def getShift(self, x, shifts, dropna = False, opt = None, **kwargs):
	# 	""" Return a shifted symbol like getShiftFromSol, but where x is the "non-shifted" symbol (pd.Series) instead of the global vector"""
	# 	m = self.shiftMaps(adj.rc_pd(x, **kwargs), shifts, **noneInit(opt, {}))
	# 	return pd.Series(x[m.values].values, index = x.index)

	# def lagMaps(self, m, lags):
	# 	return self._lagMaps(m.index, lags) if isinstance(m.index, pd.MultiIndex) else self._lagMap(m.index, lags)

	# @staticmethod
	# def _lagMaps(idx, lags):
	# 	return pd.Series(idx.set_levels([SymMaps._lagLevelMap(idx, idx.names.index(level),lag) for level,lag in lags.items()], level = lags.keys()).values, index  = idx)

	# @staticmethod
	# def _lagLevelMap(idx, levelInt, lag):
	# 	return idx.levels[levelInt].map(SymMaps._lagMap(idx.levels[levelInt], lag))

	# @staticmethod
	# def _lagMap(idx, lag):
	# 	return pd.Series(idx-lag, index = idx)

	# def rollMaps(self, m, rolls):
	# 	return self._rollMaps(m.index, rolls) if isinstance(m.index, pd.MultiIndex) else self._rollMap(m.index, rolls)

	# @staticmethod
	# def _rollMaps(idx, rolls):
	# 	return pd.Series(idx.set_levels([SymMaps._rollLevelMap(idx, idx.names.index(level), roll) for level,roll in rolls.items()], level = rolls.keys()).values, index  = idx)

	# @staticmethod
	# def _rollLevelMap(idx, levelInt, roll):
	# 	return idx.levels[levelInt].map(SymMaps._rollMap(idx.levels[levelInt], roll))

	# @staticmethod
	# def _rollMap(idx, roll):
	# 	return pd.Series(np.roll(idx, roll), index = idx)

	# def shiftMaps(self, m, shifts, **kwargs):
	# 	return self._shiftMaps(m.index, shifts, **kwargs) if isinstance(m.index, pd.MultiIndex) else self._shiftMap(m.index, shifts, **kwargs)

	# @staticmethod
	# def _shiftMaps(idx, shifts, **kwargs):
	# 	return pd.Series(idx.set_levels([SymMaps._shiftLevelMap(idx, idx.names.index(level), shift, **kwargs) for level,shift in shifts.items()], level = shifts.keys(), verify_integrity=False).values, index = idx)

	# @staticmethod
	# def _shiftLevelMap(idx, levelInt, shift, **kwargs):
	# 	idxLevel = pd.Series(idx.levels[levelInt], idx.levels[levelInt]).convert_dtypes() # allows for NA without breaking type definition
	# 	return idx.levels[levelInt].map(SymMaps._shiftOptions(idxLevel, shift, **kwargs))

	# @staticmethod
	# def _shiftMap(idx, shift, **kwargs):
	# 	return SymMaps._shiftOptions(pd.Series(idx, idx).convert_dtypes(), shift, **kwargs)

	# @staticmethod
	# def _shiftOptions(m, shift, fill_value=None, useLoc = None, useIloc = None):
	# 	if useLoc == 'nn':
	# 		return m.shift(shift, fill_value = m.iloc[shift-1 if shift>0 else shift])
	# 	elif fill_value:
	# 		return m.shift(shift, fill_value = fill_value)
	# 	elif useLoc:
	# 		return m.shift(shift, fill_value = m.loc[useLoc])
	# 	elif useIloc:
	# 		return m.shift(shift, fill_value = m.iloc[useIloc])
	# 	else:
	# 		return m.shift(shift)



from symMaps.base import *
from symMaps.simpleSys import SimpleSys
from symMaps.lpSys import AMatrix, ADict, LPSys
from symMaps.lpModels import ModelShell
# Test models:
# from symMaps.mBasic import MBasic, MBasicEmCap, MBasicRES # test


import itertools, numpy as np, pandas as pd
from collections.abc import Iterable
from pyDbs import adj, adjMultiIndex, Broadcast, Gpy, Gpy_, GpySet, GpyVariable, GpyScalar, GpyDict, SimpleDB
from pyDbs import cartesianProductIndex as CPI
_numtypes = (int,float,np.generic)


### -------- 	0: Small, auxiliary functions	-------- ###
def noneInit(x,FallBackVal):
	return FallBackVal if x is None else x

def is_iterable(arg):
	return isinstance(arg, Iterable) and not isinstance(arg, string_types)

def getIndex(symbol):
	""" Defaults to None if no index is defined. """
	if hasattr(symbol, 'index'):
		return symbol.index
	elif isinstance(symbol, pd.Index):
		return symbol
	elif not is_iterable(symbol):
		return None

def getValues(symbol):
	""" Defaults to the index, if no values are defined (e.g. if symbol is an index) """
	if isinstance(symbol, (pd.Series, pd.DataFrame, pd.Index)):
		return symbol
	elif hasattr(symbol,'v'):
		return symbol.v
	elif not is_iterable(symbol):
		return symbol

def getDomains(x):
	return [] if getIndex(x) is None else getIndex(x).names

def pdGb(x, by):
	if is_iterable(by):
		return x.groupby([k for k in x.index.names if k not in by])
	else:
		return x.groupby([k for k in x.index.names if k != by])

def pdSum(x,sumby):
	return pdGb(x, sumby).sum() if isinstance(x.index, pd.MultiIndex) else sum(x)

class Lag:
	"""
	Utility for creating lagged versions of pandas indices and Series.

	The Lag class provides static helpers that shift a pandas Index or Series
	backwards (i.e. create index/value pairs corresponding to v[index - lag]).
	It supports 1D indices and MultiIndex objects via level or dict specifications.

	Key methods
	- series(v, lag, level=None, fkeep=False, bfill='exo', exo=0):
	  Return a pd.Series with values shifted according to `lag`. Handles options:
		fkeep: keep forward-shifted indices outside original domain.
		bfill: how to backfill values not present in original series ('exo','ss', False).
		exo: numeric value used when bfill == 'exo'.
	- idx(idx, lag, level=None):
	  Return a shifted Index (or MultiIndex) corresponding to lag specification.

	Typical usage
	- Create lagged parameters or variables before adding them to linear
	  constraints in LPSys (see symMaps.lpSys.LPSys.initA_lag).
	"""
	@staticmethod
	def series(v, lag, level = None, fkeep = False, bfill = 'exo', exo = 0):
		"""Return a pd.Series shifted by `lag` on `level`.
		Index of the returned series represents positions after applying lag.
		Options:
		  - fkeep: keep forwarded indices outside original domain
		  - bfill: how to fill missing entries ('exo','ss', False)
		  - exo: numeric used when bfill == 'exo'
		"""
		vl = pd.Series(v.values, index = Lag.idx(v.index, lag, level = level))
		if fkeep is False:
			vl = vl[vl.index.isin(v.index)]
		if bfill == 'exo':
			return vl.combine_first(Lag.addExo(v, vl, exo))
		elif bfill == 'ss':
			return vl.combine_first(v)
		elif bfill is False:
			return vl

	@staticmethod
	def idx(idx, lag, level = None):
		"""Return an Index or MultiIndex shifted by `lag`.
		Dispatches to 1D or MultiIndex handlers based on idx type.
		"""
		return Lag.lagMIdx(idx, lag, level = level) if isinstance(idx, pd.MultiIndex) else Lag.lag1dIdx(idx, lag)

	@staticmethod
	def addExo(v, vl, exo):
		return pd.Series(0, index = v[~v.index.isin(vl.index)].index).add(exo)

	@staticmethod
	def lag1dIdx(idx, lag):
		""" Creates symbol[idx-lag] definition """
		return idx+lag

	@staticmethod
	def lagMIdx(idx, lags, level = None):
		return Lag.lagMIdx_dict(idx, lags) if isinstance(lags, dict) else Lag.lagMIdx_level(idx, lags, level)

	@staticmethod
	def lagMIdx_level(idx, lag, level):
		return idx.set_levels(idx.levels[idx.names.index(level)]+lag, level = level)

	@staticmethod
	def lagMIdx_dict(idx, lags):
		""" Dict maps from index level to lag """
		return idx.set_levels([idx.levels[idx.names.index(level)]+lag for level, lag in lags.items()], level = lags.keys())

class Lead(Lag):
	"""
	Lead is a thin wrapper around Lag that applies a forward shift.

	Lead inverts the sign of the lag input and then calls the Lag helpers.
	This allows callers to express leads (future indices) using the same
	interface as lags. All parameters and options are passed through to Lag.

	Example:
	  Lead.series(v, lead, fkeep=True, bfill=False)
	"""
	@staticmethod
	def series(v, lead, **kwargs):
		""" Applies Lag by inverting lead"""
		return Lag.series(v, Lead.invertLag(lead), **kwargs)

	@staticmethod
	def idx(idx, lead, **kwargs):
		""" Applies Lag by inverting lead"""
		return Lag.idx(idx, Lead.invertLag(lead), **kwargs)

	@staticmethod
	def invertLag(lag):
		return {k: -v for k,v in lag.items()} if type(lag) is dict else -lag

class Roll:
	"""
	Utilities to 'roll' (circularly shift) pandas indices and Series.

	Roll performs circular rotations of index levels. Unlike Lag, Roll does not
	create new index elements; it permutes existing ones. Methods mirror those
	of Lag but use numpy.roll semantics. Useful for circular boundary conditions
	or categorical index rotations.

	Key methods
	- series(v, roll, level=None): return a Series whose index is rolled.
	- idx(idx, roll, level=None): return rolled Index or MultiIndex.
	"""
	@staticmethod
	def series(v, roll, level = None):
		"""Return a pd.Series with its index circularly rotated by `roll` on `level`."""
		vl = pd.Series(v.values, index = Roll.idx(v.index, roll, level = level))
		return vl

	@staticmethod
	def idx(idx, roll, level = None):
		"""Return an Index or MultiIndex circularly rotated by `roll`."""
		return Roll.rollMIdx(idx, roll, level = level) if isinstance(idx, pd.MultiIndex) else Roll.roll1dIdx(idx, roll)

	@staticmethod
	def roll1dIdx(idx, roll):
		return pd.Index(np.roll(idx, roll), name = idx.name)

	@staticmethod
	def rollMIdx(idx, rolls, level = None):
		return Roll.rollMIdx_dict(idx, rolls) if isinstance(rolls, dict) else Roll.rollMIdx_level(idx, rolls, level)

	@staticmethod
	def rollMIdx_level(idx, roll, level):
		return idx.set_levels(np.roll(idx.levels[idx.names.index(level)], roll), level = level)

	@staticmethod
	def rollMIdx_dict(idx, rolls):
		return idx.set_levels([np.roll(idx.levels[idx.names.index(level)], roll) for level, roll in rolls.items()], level = rolls.keys())
	
	@staticmethod
	def invertRoll(roll):
		return {k: -v for k,v in roll.items()} if type(roll) is dict else -roll



from symMaps.base import *
from scipy import sparse
_adjF = adj.rc_pd

class AMatrix:
	""" Defines matrices in long-form. """
	def __init__(self, name, values = None, v = None, constr = None, vIdx = None, constrIdx = None):
		self.name = name # unique identifier
		self.v = v # name of variable the coefficient applies to 
		self.constr = constr # name of constraint the coefficient applies to
		self.values = values
		self.vIdx = vIdx
		self.constrIdx = constrIdx

	def __len__(self):
		return len(self.values)

	@property
	def isscalar(self):
		return all((x is None for x in (self.vIdx, self.constrIdx)))

class AMDict:
	def __init__(self, symbols = None):
		""" Simple keyword database with gpy symbols. 
			Slightly different logic than SimpleDB, as it allows for key!=self.symbols[key].name"""
		self.symbols = noneInit(symbols, {})

	def __getitem__(self,item):
		return self.symbols[item]

	def __setitem__(self,item,value):
		""" Add gpy directly with itentifier item:str, or using item:tuple. """
		if isinstance(value, AMatrix):
			self.symbols[item] = value
		else:
			self.symbols[item] = AMatrix(item, **value)

	def __call__(self, item, attr = 'value'):
		return getattr(self[item], attr)

	def __iter__(self):
		return iter(self.symbols.values())

	def __delitem__(self,item):
		del(self.symbols[item])

	def __len__(self):
		return len(self.symbols)


class LPSys:
	"""
	LPSys — a helper for building and compiling linear programs using pandas indices.

	Purpose
	- Provide a high-level, index-aware API for declaring variables, constraints
	  and their linear relationships using pandas Index / MultiIndex objects.
	- Maintain sparse definitions (series over partial domains) and compile them
	  into the dense structures required by solvers (e.g. scipy.optimize.linprog).

	Key concepts / attributes
	- self.v: dict mapping variable name -> domain (pd.Index, pd.MultiIndex or None for scalar).
	- self.eq, self.ub: dict mapping constraint name -> domain (pd.Index, pd.MultiIndex or None for scalar).
	- self.lp: dict-like store for problem components:
		- 'A_eq' / 'A_ub' : AMDictionaries of AMatrix objects (matrix coefficients).
		- 'c', 'l', 'u' etc: coefficient dictionaries (GpyDict / pandas Series).
	- self.maps: compiled mappings from pandas domains to stacked global variable indices.
	- self.len: dict of lengths for domains (used when building arrays).
	- self.out: compiled arguments (A, b, c, bounds) ready for solver.

	Important methods (high-level)
	- compileMaps(): build internal mappings (self.maps) from declared domains (self.v, self.eq, self.ub).
	- compileParams() / compileParams(...): convert self.lp parameter dicts to solver vectors/matrices.
	- initA_infer(name, series, v, constr, attr='eq'): add AMatrix using inference from series index.
	- initA_lag(...), initA_roll(...): helpers to add matrix coefficients where the variable
	  index is a lag/roll (calls symMaps.base.Lag / Lead / Roll utilities).
	- lazyA(...), lazyA_lag(...): broadcast a (possibly partial) coefficient series across
	  combined domains and add as AMatrix; variants handle lag/roll mappings before broadcasting.
	- sparse_bounds() / dense_bounds(): helpers to convert bound declarations to solver bounds.
	- unloadSol(sol): map solver solution back to pandas Series / dictionaries for variables and duals.

	Notes on lags/leads/rolls
	- This class uses Lag, Lead and Roll utilities (defined in symMaps.base) to transform
	  pandas indices/series before building matrix blocks. These utilities are expected
	  to return aligned pd.Index / pd.Series objects that match the intended constraint domains.
	- When adding lagged coefficients, the constraint domain and variable domain can be
	  different; initA_lag / lazyA_lag handle mapping between them via index transformations.

	Usage (brief example)
		self.v = {'y': pd.MultiIndex.from_product([t, i])}
		self.eq = {'vec_tx0': t[1:]}
		self.compileMaps()
		# add parameter defined on y domain
		par = pd.Series(..., index=self.v['y'])
		# add a lagged coefficient that aligns par[t-1,i] with constraint vec_tx0[t]
		self.initA_lag('vec_tx0_y', series=par, v='y', constr='vec_tx0', attr='eq',
					   lag=1, level='t', fkeep=False, bfill=False)

	Implementation details
	- Internally uses lightweight AMatrix objects and AMDict containers to keep the
	  model sparse until compileParams() is called.
	- Many methods accept scalar, 1d or nd index inputs and dispatch based on
	  pandas Index vs MultiIndex types.

	See also
	- symMaps.base.Lag, Lead, Roll — helpers for index/series transformations.
	- AMatrix, AMDict definitions in this module (below) — low-level matrix block abstractions.
	"""
	def __init__(self, db = None, v = None, eq = None, ub = None, scalarDual = True):
		self.db = noneInit(db, SimpleDB()) # database
		self.v = noneInit(v, {}) # used for establishing global domains for variables.
		self.eq = noneInit(eq, {}) # used for domains eq. constraints.
		self.ub = noneInit(ub, {}) # used for domains ineq. constraints.
		# self.lp = {k: GpyDict() for k in ('c','b_eq','b_ub','l','u')} | {k: GpyConstrDict() for k in ('A_ub', 'A_eq')} # collect dense parameters here.
		self.lp = {k: GpyDict() for k in ('c','b_eq','b_ub','l','u')} | {k: AMDict() for k in ('A_ub', 'A_eq')} # collect dense parameters here.
		self.out = {k: None for k in ('c','A_ub', 'b_ub', 'A_eq','b_eq','bounds')} # store final objects here.
		self.len = dict.fromkeys(('v','eq','ub'))
		self.maps = dict.fromkeys(('v','eq','ub'))
		self.scalarDual = scalarDual # if True this checks for instances where lower bound = upper bound. The shadow value is then added to the upper bound. 

	def __getitem__(self, item):
		return self.maps['v'][item]

	def __call__(self, x, name, **kwargs):
		""" Subset x with index of variable 'name' using linear indexing from self.maps."""
		return x[self[name].values]

	def get(self, x, name, **kwargs):
		""" __call__ method, but returned as pandas object (or scalar)"""
		return x[self[name][0]] if self.v[name] is None else pd.Series(x[self[name]], index = self[name].index, name = name)

	def x0(self, attr = 'v', fill_value = 0):
		return np.concatenate([self.xFromDb(name, attr = attr, fill_value = fill_value) for name in self.v])

	def xFromDb(self, name, attr = 'v', fill_value = 0):
		return self.db[name].array(attr = attr) if name in self.db.symbols else np.full(len(self.v[name]), fill_value)

	def unloadSol(self, sol, pfeq = None, pfub = None, pfl = None, pfu = None, **kwargs):
		return (self.unloadSolX(sol) | self.unloadSolDualEq(sol, prefix = pfeq) | self.unloadSolDualUb(sol, prefix = pfub) | self.undloadSolDualLower(sol, prefix=pfl) | self.undloadSolDualUpper(sol, prefix=pfu))

	def unloadSolX(self, sol):
		""" Map solution array to dictionary of pd.Series (or scalar) """
		return {k: self.get(sol['x'], k) for k in self.v}

	# Get shadow values on constraints:
	def getDualConstr(self, x, name, attr = 'eq', **kwargs):
		return x[self.maps[attr][name][0]] if getattr(self,attr)[name] is None else pd.Series(x[self.maps[attr][name].values], index = self.maps[attr][name].index, name = name)

	def unloadSolDualEq(self, sol, prefix = None, **kwargs):
		prefix = noneInit(prefix, 'λeq_')
		return {f'{prefix}{k}': self.getDualConstr(sol['eqlin']['marginals'], k, 'eq') for k in self.eq}

	def unloadSolDualUb(self, sol, prefix = None, **kwargs):
		prefix = noneInit(prefix, 'λub_')
		return {f'{prefix}{k}': self.getDualConstr(sol['ineqlin']['marginals'], k, 'ub') for k in self.ub}

	# Get shadow values on bounds:
	def getDualBound(self, x, name, **kwargs):
		return x[self.maps['v'][name][0]] if self.v[name] is None else pd.Series(x[self.maps['v'][name].values], index = self.maps['v'][name].index, name = name)

	def undloadSolDualLower(self, sol, prefix = None, **kwargs):
		vals = self.scalarDualLower(sol) if self.scalarDual else sol['lower']['marginals']
		prefix = noneInit(prefix, 'λl_')
		return {f'{prefix}{k}': self.getDualBound(vals, k) for k in self.v}

	def undloadSolDualUpper(self, sol, prefix = None, **kwargs):
		vals = self.scalarDualUpper(sol) if self.scalarDual else sol['upper']['marginals']
		prefix = noneInit(prefix, 'λu_')
		return {f'{prefix}{k}': self.getDualBound(vals, k) for k in self.v}

	# Adjust for shadow values when lower and upper bounds coincide - set lower bound to zero and add to upper bound:
	def scalarDualLower(self, sol):
		return np.where(self.out['bounds'][:,0]==self.out['bounds'][:,1], 0, sol['lower']['marginals'])

	def scalarDualUpper(self, sol):
		return np.where(self.out['bounds'][:,0]==self.out['bounds'][:,1], np.add(sol['lower']['marginals'], sol['upper']['marginals']), sol['upper']['marginals'])

	def compile(self, **kwargs):
		self.compileMaps()
		self.compileParams()
		return self.out 

	def compileMaps(self):
		"""
		Build internal mappings from declared pandas domains to stacked global indices.

		Purpose
		- Inspect self.v, self.ub, self.eq and create self.maps entries that map each pandas
		  Index / MultiIndex position to an integer position in the stacked variable
		  vector used by the solver.

		Parameters
		- None (reads declarations from self.v, self.eq, and self.ub)

		Returns
		- None (updates self.maps and self.len in-place)

		Notes
		- Must be called after declaring variables / constraints and before compiling
		  parameters or building solver arrays.
		"""
		[self.compileMaps1d(attr) for attr in ('v','eq','ub')];

	def compileParams(self):
		"""
		Compile sparse parameter declarations into solver-ready arrays.

		Purpose
		- Convert entries stored in self.lp (c, l, u, A_eq, A_ub, etc.) into the
		  dense / sparse numpy arrays and vectors required by the solver and store
		  the result in self.out.

		Parameters
		- None (reads from self.lp (c, l, u, A_eq, A_ub, etc.))

		Returns
		- None (updates self.out in-place)

		Notes
		- Assumes compileMaps() has already been run and self.maps is available.
		- scipy.optimize.linprog only works with sparse A_eq, A_ub (as of version 1.16). 
		- Sparse vectors are implemented below, but we call the dense versions here (sparse matrices)
		"""
		self.out['c'] = self.dense_c()
		self.out['b_eq'] = self.dense_b_eq()
		self.out['b_ub'] = self.dense_b_ub()
		self.out['bounds'] = self.dense_bounds()
		self.out['A_eq'] = self.sparse_A_eq()
		self.out['A_ub'] = self.sparse_A_ub()
		return self.out

	def compileMaps1d(self, attr):
		keys, vals = list(getattr(self,attr).keys()), list(getattr(self, attr).values())
		steps = np.array([0]+[1 if x is None else len(x) for x in vals]).cumsum()
		self.len[attr] = steps[-1]
		self.maps[attr] = {keys[i]: pd.Series(range(steps[i], steps[i+1]), index = getIndex(vals[i])) for i in range(len(keys))}

	### METHODS TO ADD COEFFICIENTS A_eq, A_ub from self.lp['A_eq'], self.lp['A_ub']:
	def getVIdx(self, a):
		if a.isscalar:
			return self.getIdx_scalar(a.v, 'v')
		elif a.vIdx is None:
			return self.getIdx_scalarAttr(a, a.v, 'v')
		else:
			return self.getIdx_vecAttr(a.vIdx, a.v, 'v')

	def getConstrIdx(self, a, attr):
		""" attr ∈ {'eq','ub'}"""
		if a.isscalar:
			return self.getIdx_scalar(a.constr, attr)
		elif a.constrIdx is None:
			return self.getIdx_scalarAttr(a, a.constr, attr)
		else:
			return self.getIdx_vecAttr(a.constrIdx, a.constr, attr)

	def getIdx_scalar(self, name, attr):
		return self.maps[attr][name][0]
	def getIdx_scalarAttr(self, a, name, attr):
		return np.full(len(a), self.getIdx_scalar(name, attr))
	def getIdx_vecAttr(self, idx, name, attr):
		return self.maps[attr][name][idx].values

	def getGlobalIdx1d(self, name, idx, attr):
		return self.maps[attr][name][0] if idx is None else self.maps[attr][name][idx].values

	def sparse_A_ub(self):
		if len(self.lp['A_ub'])>0:
			return sparse.coo_array((np.hstack([sym.values for sym in self.lp['A_ub']]), 
									(np.hstack([self.getConstrIdx(sym, 'ub') for sym in self.lp['A_ub']]), 
									 np.hstack([self.getVIdx(sym) for sym in self.lp['A_ub']]))), shape = (self.len['ub'], self.len['v']))
		else:
			return sparse.coo_array(([], ([],[])), shape = (self.len['ub'], self.len['v']))

	def sparse_A_eq(self):
		if len(self.lp['A_eq'])>0:
			return sparse.coo_array((np.hstack([sym.values for sym in self.lp['A_eq']]), 
									(np.hstack([self.getConstrIdx(sym, 'eq') for sym in self.lp['A_eq']]), 
									 np.hstack([self.getVIdx(sym) for sym in self.lp['A_eq']]))), shape = (self.len['eq'], self.len['v']))
		else:
			return sparse.coo_array(([], ([],[])), shape = (self.len['eq'], self.len['v']))

	# Methods for initializing A-blocks:
	def initA_infer(self, name, series = None, v = None, constr = None, attr = None):
		"""
		Add an AMatrix block by inferring vIdx and constrIdx from `series`.

		Purpose
		- Create and register an AMatrix using the provided pd.Series. The method
		  infers the variable-domain (vIdx) and constraint-domain (constrIdx) from
		  the series.index and aligns those with declared domains for `v` and `constr`.

		Parameters
		- name: unique identifier for the AMatrix block
		- series: pd.Series containing coefficients over some subset of the combined domain
		- v: variable name (must exist in self.v)
		- constr: constraint name (must exist in self.eq)
		- attr: 'eq' or 'ub' (which AMDict to register under)
		- kwargs: additional options forwarded to AMatrix construction (broadcasting, etc.)

		Returns
		- None (registers AMatrix in self.lp['A_<attr>'])
		"""
		self.lp[f'A_{attr}'][name] = AMatrix(name, values = series.values, v = v, constr= constr, 
													vIdx  = self.inferIdx(series, v, 'v'), 
													constrIdx = self.inferIdx(series, constr, attr))

	def inferIdx(self, series, name, attr):
		if getattr(self,attr)[name] is None:
			return None
		else:
			return self.inferIdxFromIdx(series.index, name, attr)
	def inferIdxFromIdx(self, idx, name, attr):
		if isinstance(getattr(self,attr)[name], pd.MultiIndex):
			return self.inferNdIdx(idx, name, attr)
		else:
			return self.infer1dIdx(idx, name, attr)
	def inferNdIdx(self, idx, name, attr):
		return self.maps[attr][name][pd.MultiIndex.from_arrays([idx.get_level_values(n) for n in self.maps[attr][name].index.names])].index
	def infer1dIdx(self, idx, name, attr):
		return self.maps[attr][name][idx.get_level_values(self.maps[attr][name].index.name)].index


	# Methods to initialize vectors c, b_eq, b_ub, l, u:
	def lazyV(self, name, series = None, v = None, attr = None, how = 'inner'):
		""" series: pd.Series or scalar. Domains has to be consistent with variable domains. attr ∈ {'v','l','u'}"""
		if self.v[v] is None:
			self.lp[attr][(name, v)] = series # scalar index
		else:
			self.lp[attr][(name, v)] = Broadcast.valuesToIdx(series, self.v[v], how = how)

	def lazyC(self, name, **kwargs):
		self.lazyV(name, attr = 'c', **kwargs)
	def lazyL(self, name, **kwargs):
		self.lazyV(name, attr = 'l', **kwargs)
	def lazyU(self, name, **kwargs):
		self.lazyV(name, attr = 'u', **kwargs)

	def lazyB(self, name, series = None, constr = None, attr = None, how = 'inner'):
		""" series: pd.Series or scalar. Domains has to be consistent with constraint domains. """
		if getattr(self, attr)[constr] is None:
			self.lp[f'b_{attr}'][(name, constr)] = series # scalar index
		else:
			self.lp[f'b_{attr}'][(name, constr)] = Broadcast.valuesToIdx(series, getattr(self, attr)[constr], how = how)


	def initA_lag(self, name, series = None, v = None, constr = None, attr = None, lag = None, level = None, fkeep = False, bfill = False, **kwargs):
		"""
		Add lagged linear constraint. 
		name: Unique identifier added to self.lp[f'A_{attr}'].
		series: pd.Series, coefficients to apply to lagged variable.
		lag: int or dict with, lag/lead specification (see Lag.series).
		level: None or str, specifies level of inddex to lag if type(lag) is int.
		v: str, name of variable the coefficient applies to in linear constraint.
		attr: 'eq' or 'ub', constraint type.
		fkeep: bool. If True --> keep indices that are leaded beyond original index. 
		bfill: False or in ('exo','ss'). Specifies how to backfill values that are not in original parameter. If False --> we do not fill in any values.
		kwargs: passed to Lag.series.
		"""
		laggedSeries = Lag.series(series, lag, level = level, fkeep = fkeep, bfill = bfill, **kwargs)
		self._addLaggedA(name, laggedSeries, v = v, constr = constr, attr= attr, lag = lag, level = level)

	def _addLaggedA(self, name, laggedSeries, v = None, constr = None, attr = None, lag = None, level = None):
		self.lp[f'A_{attr}'][name] = AMatrix(name, values = laggedSeries.values, v = v, constr = constr, 
													vIdx = self.inferIdxFromIdx(Lead.idx(laggedSeries.index, lag, level = level) , v, 'v'),
													constrIdx = self.inferIdx(laggedSeries, constr, attr))

	def initA_roll(self, name, series = None, v = None, constr = None, attr = None, roll = None, level = None, **kwargs):
		""" Similar to initA_lag, but with circular index reference. """
		rollSeries = Roll.series(series, roll, level = level,**kwargs)
		self._addRolledA(name, rollSeries, v = v, constr = constr, attr = attr, roll = roll, level = level)

	def _addRolledA(self, name, rollSeries, v = None, constr = None, attr= None, roll = None, level = None):
		self.lp[f'A_{attr}'][name] = AMatrix(name, values = rollSeries.values, v = v, constr = constr, 
													vIdx = self.inferIdxFromIdx(Roll.idx(rollSeries.index, Roll.invertRoll(roll), level = level) , v, 'v'),
													constrIdx = self.inferIdx(rollSeries, constr, attr))		


	def lazyA_lag(self, name, series = None, v = None, constr = None, attr = None, constrIdx = None, vIdx = None, lag = None, **kwargs):
		""" Same inputs as self.lazyA with added lag method on the variable index. Note: only works when constraint and variable are pd.Indices (not scalars)"""
		vIdx, constrIdx = noneInit(vIdx, self.v[v]), noneInit(constrIdx, getattr(self,attr)[constr])
		vSeries = Broadcast.valuesToIdx(series, vIdx) # broadcast to full variable index.
		lSeries = Lag.series(vSeries, lag, fkeep = True, bfill = False) # lag series mapped to variable index.
		fIdx = Broadcast.idx(constrIdx, lSeries.index) # broadcast full index.
		flSeries = Broadcast.valuesToIdx(lSeries, fIdx) # full lagged series.
		self._addLaggedA(name, flSeries, v = v, constr = constr, attr= attr, lag = lag)

	def lazyA_roll(self, name, series = None, v = None, constr = None, attr = None, constrIdx = None, vIdx = None, roll = None, **kwargs):
		""" Same inputs as self.lazyA with added lag method on the variable index. Note: only works when constraint and variable are pd.Indices (not scalars)"""
		vIdx, constrIdx = noneInit(vIdx, self.v[v]), noneInit(constrIdx, getattr(self,attr)[constr])
		vSeries = Broadcast.valuesToIdx(series, vIdx) # broadcast to full variable index.
		lSeries = Roll.series(vSeries, roll) # roll series mapped to variable index.
		fIdx = Broadcast.idx(constrIdx, lSeries.index) # broadcast full index.
		flSeries = Broadcast.valuesToIdx(lSeries, fIdx) # full lagged series.
		self._addRolledA(name, flSeries, v = v, constr = constr, attr= attr, lag = lag)

	def lazyA(self, name, series = None, v = None, constr = None, attr = None, constrIdx = None, vIdx = None, **kwargs):
		""" 
		Broadcast a (possibly partial) coefficient series over combined domains and add as AMatrix.

		Purpose
		- Given a pd.Series that covers some subset of the combined constraint-variable
		  domain, broadcast it to the full combined index and register an AMatrix block.

		Parameters
		- name: str, unique identifier added in self.lp[f'A_{attr}'].
		- series: pd.Series or scalar, coefficients defined over a subset of the cartesian product of vIdx and constrIdx.
		- v: str, name of variable accessed in self.v.
		- constr: str, name of constraint accessed in getattr(self, attr).
		- attr: 'eq' or 'ub' , indicates type of constraint.
		- constrIdx: None or pd.Index. Domain of constraint (subset of the full domain in getattr(self, attr)[constr]). If None --> use full domain.
		- vIdx: None or pd.Index. Domain of variable (subset of the full domain in self.v[v]). If None --> use full domain.
		"""
		vIdx, constrIdx = noneInit(vIdx, self.v[v]), noneInit(constrIdx, getattr(self,attr)[constr])
		if all((isinstance(x, pd.Index) for x in (vIdx, constrIdx))):
			fIdx = Broadcast.idx(vIdx, constrIdx)
			fSeries = Broadcast.valuesToIdx(series, fIdx)
			values, vIdx, constrIdx = fSeries.values, self.inferIdx(fSeries, v, 'v'), self.inferIdx(fSeries, constr, attr)
		elif constrIdx is None: 
			fSeries = Broadcast.valuesToIdx(series, vIdx)
			values, vIdx = fSeries.values, fSeries.index
		elif vIdx is None:
			fSeries = Broadcast.valuesToIdx(series, constrIdx) # Map coefficient to constraint index
			values, constrIdx = fSeries.values, fSeries.index
		else: 
			values = series # scalar 
		self.lp[f'A_{attr}'][name] = AMatrix(name, values = values, v = v, constr = constr, vIdx = vIdx, constrIdx = constrIdx)


	# NOTE: Sparse vectors are not yet implemented in scipy optimize 1.16 - only matrices A_eq, A_ub from above. 
	def dense_bounds(self):
		bounds = np.full((self.len['v'], 2), np.nan)
		if len(self.lp['l'])>0:
			bounds[self._rowsStack('l','v'), 0] = self._dataStack('l')
		if len(self.lp['u'])>0:
			bounds[self._rowsStack('u','v'), 1] = self._dataStack('u')
		return bounds
	def sparse_bounds(self):
		if len(self.lp['l'])>0:
			l = self._dataStack('l')
			lRowIdx = self._rowsStack('l','v')
		else:
			l,lRowIdx = [], []
		if len(self.lp['u'])>0:
			u = self._dataStack('u')
			uRowIdx = self._rowsStack('u','v')
		else:
			u,uRowIdx = [], []
		return sparse.coo_array((np.hstack([l,u]), (np.hstack([lRowIdx, uRowIdx]), np.hstack([np.zeros(len(l)), np.ones(len(u))]))), shape = (self.len['v'], 2))

	def dense_c(self):
		c = np.zeros(self.len['v'])
		if len(self.lp['c'])>0:
			c[self._rowsStack('c','v')] = self._dataStack('c')
		return c
	def sparse_c(self):
		if len(self.lp['c'])>0:
			return sparse.coo_array((self._dataStack('c'), (self._rowsStack('c','v'),)), shape = (self.len['v'],))
		else:
			return sparse.coo_array(([], ([],)), shape = (self.len['v'],))

	def dense_b_eq(self):
		b = np.zeros(self.len['eq'])
		if len(self.lp['b_eq'])>0:
			b[self._rowsStack('b_eq','eq')] = self._dataStack('b_eq')
		return b
	def sparse_b_eq(self):
		if len(self.lp['b_eq'])>0:
			return sparse.coo_array((self._dataStack('b_eq'), (self._rowsStack('b_eq','eq'), )), shape = (self.len['eq'],))
		else:
			return sparse.coo_array(([], ([],)), shape = (self.len['eq'],))

	def dense_b_ub(self):
		b = np.zeros(self.len['ub'])
		if len(self.lp['b_ub'])>0:
			b[self._rowsStack('b_ub','ub')] = self._dataStack('b_ub')
		return b
	def sparse_b_ub(self):
		if len(self.lp['b_ub'])>0:
			return sparse.coo_array((self._dataStack('b_ub'), (self._rowsStack('b_ub','ub'), )), shape = (self.len['ub'],))
		else:
			return sparse.coo_array(([], ([],)), shape = (self.len['ub'],))

	def _dataStack(self,key):
		return np.hstack([sym.array() for sym in self.lp[key]])
	def _rowsStack(self, key, attr):
		return np.hstack([self.getGlobalIdx1d(sym.name, sym.index, attr) for sym in self.lp[key]])




from symMaps.base import *
from symMaps.lpSys import AMatrix, AMDict, LPSys
from symMaps.lpModels import ModelShell
from scipy import sparse, optimize
_adjF = adj.rc_pd

# A few basic functions for the energy models:
def fuelCost(db):
	""" 
	 - 'pFuel' (fuel price) is defined over 'idxF' (fuel index). Default unit: €/GJ.
	 - 'uEm' (emission intensity) is defined over 'idxF','idxEm' (emission index). Default unit: Ton emission/GJ fuel input. 
	 - 'taxEm' (tax on emissions) is defined over 'idxEm' (emission index). Default unit: €/ton emission output.
	"""
	return db('pFuel').add(pdSum((db('uEm') * db('taxEm')).dropna(), 'idxEm'), fill_value=0)

def mc(db):
	""" 
	- 'uFuel': Fuelmix is defined over 'idxF', 'idxGen' (generator index). Default unit: GJ input/GJ output.
	- 'VOM': variable operating and maintenance costs is defined over 'idxGen'. Default unit: €/GJ output.
	"""
	return pdSum((db('uFuel') * fuelCost(db)).dropna(), 'idxF').add(db['VOM'], fill_value=0)

def fuelConsumption(db, sumOver='idxGen'):
	"""
	- 'generation': dispatched energy defined over 'idxGen'. Default unit: GJ.
	- 'uFuel': Fuelmix is defined over 'idxF', 'idxGen' (generator index). Default unit: GJ input/GJ output.
	"""
	return pdSum((db('generation') * db('uFuel')).dropna(), sumOver)

def emissionsFuel(db, sumOver='idxF'):
	""" 
	- 'fuelCons': fuel input defined over 'idxF'. Default unit: GJ. 
	 - 'uEm' (emission intensity) is defined over 'idxF','idxEm' (emission index). Default unit: Ton emission/GJ fuel input. 
	"""
	return pdSum((db('fuelCons') * db('uEm')).dropna(), sumOver)

def plantEmissionIntensity(db):
	""" 
	- 'uFuel': Fuelmix is defined over 'idxF', 'idxGen' (generator index). Default unit: GJ input/GJ output.
	 - 'uEm' (emission intensity) is defined over 'idxF','idxEm' (emission index). Default unit: Ton emission/GJ fuel input. 
	"""
	return (db('uFuel') * db('uEm')).groupby('idxGen').sum()

class MBasic(ModelShell):
	def compile(self, **kwargs):
		""" Compile model """
		self.initArgs()
		return self.sys.compile(**kwargs)

	def initArgs(self):
		[getattr(self, f'initArg_{k}')() for k in ('v','eq','ub') if hasattr(self, k)]; # specify domains for variables and equations
		[getattr(self, f'initArgsV_{k}')() for k in self.v]; # specify c,l,u for all variables
		[getattr(self, f'initArgsEq_{k}')() for k in self.eq]; # add A_eq and b_eq.
		[getattr(self, f'initArgsUb_{k}')() for k in self.ub]; # add A_ub and b_ub.

	def initArgs_v(self):
		""" self.sys.v dictionary"""
		self.sys.v.update({'generation': self.db['idxGen'],'demand': self.db['idxCons']})

	def initArgs_eq(self):
		""" self.sys.eq dictionary"""
		self.sys.eq.update({'equilibrium': None})

	def initArgsV_generation(self):
		self.sys.lp['c'][('mc', 'generation')] = self.db['mc'] # assumes that mc is defined over index 'idxGen'.
		self.sys.lp['u'][('cap', 'generation')] = self.db['genCap'] # assumes that genCap is defined over index 'idxGen'.

	def initArgsV_demand(self):
		self.sys.lp['c'][('mwp', 'demand')] = -self.db['mwp'] # assumes that mwp is defined over index 'idxCons'.
		self.sys.lp['u'][('cap', 'demand')] = self.db['load'] # assumes that load is defined over index 'idxCons'.

	def initArgsEq_equilibrium(self):
		self.sys.lazyA('eq2Gen', series = 1,  v = 'generation', constr = 'equilibrium',attr='eq')
		self.sys.lazyA('eq2Dem', series = -1, v = 'demand', constr = 'equilibrium',attr='eq')

	def postSolve(self, sol, **kwargs):
		super().postSolve(sol)
		self.db['surplus'] = -solution['fun']
		self.db['fuelCons'] = fuelConsumption(self.db)
		self.db['emissions'] = emissionsFuel(self.db)


class MBasicEmCap(MBasic):
	# Add additional constraint:
	def initArgs_ub(self):
		""" self.sys.ub dictionary"""
		self.sys.ub.update({'emCap': None})

	def initArgsUb_emCap(self):
		self.sys.lazyA('emCap2Gen', series = plantEmissionIntensity(self.db),  v = 'generation', constr = 'emCap',attr='ub')
		self.sys.lp['b_ub'] = self.db('CO2Cap')


class MBasicRES(MBasic):

	def RESGenIdx(self, CO2Idx = 'CO2'):
		""" Subset of idxGen that is considered Renewable Energy based on emission intensities """
		s = self.db('uFuel') * self.db('uEm').xs(CO2Idx,level='idxEm')
		return s[s <= 0].index

	def initArgs_ub(self):
		""" self.sys.ub dictionary"""
		self.sys.ub.update({'RES': None})

	def initArgsUb_emCap(self):
		self.sys.lazyA('RES2Gen', series = pd.Series(-1, index = self.RESGenIdx()),  v = 'generation', constr = 'RES',attr='ub')
		self.sys.lazyA('RES2Dem', series = pd.Series(self.db('RESCap'), index = self.RESGenIdx()),  v = 'generation', constr = 'RES',attr='ub')



import setuptools

with open("README.md", "r") as file:
  long_description = file.read()

setuptools.setup(
  name="symMaps",
  version="0.0.1",
  author="Rasmus K. Skjødt Berg",
  author_email="rasmus.kehlet.berg@econ.ku.dk",
  description="System of mappings to navigate stacked system of equations defined over pandas indices",
  long_description=long_description,
  long_description_content_type="text/markdown",
  url="https://github.com/ChampionApe/symMaps",
  packages=setuptools.find_packages(),
  classifiers=[
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
  ],
  python_requires='>=3.11',
  install_requires=["pandas", "scipy","pyDbs"],
)


from symMaps.base import *
from scipy import sparse, optimize
from symMaps.lpSys import AMatrix, AMDict, LPSys
_adjF = adj.rc_pd

# To do: Loop through and change grid values in already compiled models.
class ModelShell:
	def __init__(self, db = None, sys = None, scalarDual = True, computeDual = True, solOptions = None, **kwargs):
		self.db = noneInit(db, SimpleDB())
		self.computeDual = computeDual
		self.scalarDual = scalarDual
		self.solOptions = self.defaultSolOptions | noneInit(solOptions, {}) # passed to optimize.linprog
		self.sys = noneInit(sys, LPSys(db = self.db, scalarDual = self.scalarDual))

	@property
	def defaultSolOptions(self):
		return {'method': 'highs', 'x0': None}
	
	def x0(self, attr = 'v', fill_value = 0):
		return self.sys.x0(attr = attr, fill_value = fill_value)

	def solve(self, **kwargs):
		""" Assumes that self.sys is compiled. """
		return optimize.linprog(**self.sys.out, **self.solOptions)

	def postSolve(self, sol, **kwargs):
		assert sol['status'] == 0, "scipy.optimize.linprog did not yield solution status == 0. Check output from self.solve()."
		dictSol = self.sys.unloadSol(sol) if self.computeDual else self.sys.unloadSolX(sol)
		[self.db.__setitem__(k,v) for k,v in dictSol.items()];
